@@ -309,14 +309,16 @@ class SQLAlchemyDatabase:
         data: pl.DataFrame,
         mode: t.Literal["RAISE", "IGNORE", "UPDATE"],
         *,
-        threshold_for_copy: int = 100000,  # temporarily do not use copy
+        threshold_for_copy: int = 1_000_000,  # temporarily do not use copy
         returning_id: bool = False,
         parallel_workers: int = 1,
     ) -> pl.DataFrame | None:
+        
         """
         Validate a Polars DataFrame against the model schema, then delegate
         to the bulk-insert implementation.
         """
+
         self.check_mode(mode)
         # 1. Basic type & column-name checks
         if not isinstance(data, pl.DataFrame):
@@ -555,7 +557,7 @@ class SQLAlchemyDatabase:
                 if not key_rows:
                     continue
                 temp_table_name = f"temp_keys_{i}"
-                session.execute(text(f'DROP TABLE IF EXISTS "{temp_table_name}"'))
+                # session.execute(text(f'DROP TABLE IF EXISTS "{temp_table_name}"'))
                 session.execute(text(f'CREATE TEMP TABLE "{temp_table_name}" ({col_defs_sql}) ON COMMIT DROP'))
                 col_objs = []
                 for col in p_unique_key_columns:
@@ -576,6 +578,13 @@ class SQLAlchemyDatabase:
                     ids.append(row_dict)
                 if hasattr(progress_bar, "update"):
                     progress_bar.update(len(key_rows))
+                
+                # Commit every 100 iterations to avoid hitting connection limits
+                if (i + 1) % 100 == 0:
+                    session.commit()
+            
+            # Final commit for any remaining uncommitted work
+            session.commit()
         return ids
 
     def handle_insertion_multi_line_copy(
@@ -622,6 +631,10 @@ class SQLAlchemyDatabase:
             )
 
         all_id_maps: list[pl.DataFrame] = []
+        
+        # Performance tracking
+        total_times = {"csv_gen": 0, "copy_data": 0, "insert": 0, "fetch_ids": 0, "truncate": 0}
+        batch_count = 0
 
         # Use single connection for all batches
         raw_conn = self.engine.raw_connection()
@@ -657,25 +670,38 @@ class SQLAlchemyDatabase:
                 batch = data.slice(offset, chunk_size)
                 if batch.height == 0:
                     continue
+                
+                batch_count += 1
 
-                # 1) Clear temp table (much faster than CREATE/DROP)
+                # 1) Time truncate operation
+                start = time.time()
                 cursor.execute(f'TRUNCATE "{temp_table}"')
+                total_times["truncate"] += time.time() - start
 
-                # 2) COPY data into staging
+                # 2) Time CSV generation
+                start = time.time()
                 buf = io.StringIO()
                 batch.write_csv(buf)
                 buf.seek(0)
-                cursor.copy_expert(f'COPY "{temp_table}" ({non_id_col_list}) FROM STDIN WITH CSV HEADER', buf)
+                total_times["csv_gen"] += time.time() - start
 
-                # 3) INSERT FROM staging into real table
+                # 3) Time COPY operation
+                start = time.time()
+                cursor.copy_expert(f'COPY "{temp_table}" ({non_id_col_list}) FROM STDIN WITH CSV HEADER', buf)
+                total_times["copy_data"] += time.time() - start
+
+                # 4) Time INSERT operation
+                start = time.time()
                 insert_sql = (
                     f'INSERT INTO "{table}" ({non_id_col_list}) SELECT {non_id_col_list} FROM "{temp_table}" '
                     + on_conflict
                 )
                 cursor.execute(insert_sql)
+                total_times["insert"] += time.time() - start
 
-                # 4) Optionally retrieve IDs by joining via COPY + Polars
+                # 5) Time ID fetching
                 if returning_id:
+                    start = time.time()
                     buf_out = io.StringIO()
                     # Build SELECT clause with explicit aliases
                     select_list = ", ".join(['r."ID" AS "ID"', *[f't."{c}" AS "{c}"' for c in pks]])
@@ -684,6 +710,7 @@ class SQLAlchemyDatabase:
                     buf_out.seek(0)
                     df_ids = pl.read_csv(buf_out)
                     all_id_maps.append(df_ids)
+                    total_times["fetch_ids"] += time.time() - start
 
                 # Update progress bar with number of rows processed
                 if show_progress:
@@ -696,6 +723,20 @@ class SQLAlchemyDatabase:
             cursor.close()
             raw_conn.close()
 
+        # Log performance breakdown
+        total_time = sum(total_times.values())
+        total_rows = data.height
+        print(
+            f"COPY Performance Breakdown for {total_rows:,} rows in {batch_count} batches:\n"
+            f"  CSV generation: {total_times['csv_gen']:.3f}s ({total_times['csv_gen']/total_time*100:.1f}%)\n"
+            f"  COPY to temp:   {total_times['copy_data']:.3f}s ({total_times['copy_data']/total_time*100:.1f}%)\n"
+            f"  INSERT upsert:  {total_times['insert']:.3f}s ({total_times['insert']/total_time*100:.1f}%)\n"
+            f"  Fetch IDs:      {total_times['fetch_ids']:.3f}s ({total_times['fetch_ids']/total_time*100:.1f}%)\n"
+            f"  Truncate:       {total_times['truncate']:.3f}s ({total_times['truncate']/total_time*100:.1f}%)\n"
+            f"  Total time:     {total_time:.3f}s\n"
+            f"  Overall rate:   {total_rows/total_time:.0f} rows/sec"
+        )
+
         if returning_id:
             all_dfs = pl.concat(all_id_maps)
             return all_dfs
@@ -786,19 +827,43 @@ class SQLAlchemyDatabase:
             )
 
         all_id_maps = []
+        all_worker_times = []
 
-        # Use multiprocessing Pool
+        # Use multiprocessing Pool with imap_unordered for real-time updates
         with mp.Pool(processes=num_workers) as pool:
-            # Submit all chunks to workers
-            results = pool.map(worker_func, chunks)
-
-            # Collect results and update progress
-            for result in results:
-                batch_size, id_df = result
+            # Submit all chunks to workers and process results as they complete
+            for result in pool.imap_unordered(worker_func, chunks):
+                batch_size, id_df, worker_times = result
+                all_worker_times.append(worker_times)
                 if show_progress:
                     progress_bar.update(batch_size)
                 if returning_id and id_df is not None:
                     all_id_maps.append(id_df)
+        
+        if show_progress:
+            progress_bar.close()
+
+        # Aggregate and log performance statistics
+        if all_worker_times:
+            # Calculate averages across all workers
+            avg_times = {}
+            for key in all_worker_times[0].keys():
+                avg_times[key] = sum(worker[key] for worker in all_worker_times) / len(all_worker_times)
+            
+            total_avg_time = sum(avg_times.values())
+            total_rows = data.height
+            
+            logger.info(
+                f"Parallel COPY Performance Breakdown (average per worker) for {total_rows:,} rows with {num_workers} workers:\n"
+                f"  CSV generation:   {avg_times['csv_gen']:.3f}s ({avg_times['csv_gen']/total_avg_time*100:.1f}%)\n"
+                f"  COPY to temp:     {avg_times['copy_data']:.3f}s ({avg_times['copy_data']/total_avg_time*100:.1f}%)\n"
+                f"  INSERT upsert:    {avg_times['insert']:.3f}s ({avg_times['insert']/total_avg_time*100:.1f}%)\n"
+                f"  Fetch IDs:        {avg_times['fetch_ids']:.3f}s ({avg_times['fetch_ids']/total_avg_time*100:.1f}%)\n"
+                f"  Temp table ops:   {avg_times['temp_table_ops']:.3f}s ({avg_times['temp_table_ops']/total_avg_time*100:.1f}%)\n"
+                f"  Avg worker time:  {total_avg_time:.3f}s\n"
+                f"  Workers used:     {len(all_worker_times)}\n"
+                f"  Parallel efficiency: {len(all_worker_times) * total_avg_time:.1f}s total worker time"
+            )
 
         if returning_id and all_id_maps:
             return pl.concat(all_id_maps)
@@ -881,17 +946,20 @@ def _copy_worker_static(
     table: str,
     cols: list[str],
     returning_id: bool,
-) -> tuple[int, pl.DataFrame | None]:
+) -> tuple[int, pl.DataFrame | None, dict[str, float]]:
     """
     Static worker function for parallel COPY operations.
     This function recreates its own database connection from the connection URL.
 
     Returns:
-        tuple: (batch_size, id_dataframe_or_none)
+        tuple: (batch_size, id_dataframe_or_none, timing_stats)
 
     """
     offset, batch = chunk_data
     batch_size = batch.height
+    
+    # Performance tracking
+    worker_times = {"csv_gen": 0, "copy_data": 0, "insert": 0, "fetch_ids": 0, "temp_table_ops": 0}
 
     # Create new engine and connection for this worker
     worker_engine = create_engine(connection_url)
@@ -921,26 +989,37 @@ def _copy_worker_static(
         non_id_cols = [c for c in cols if c.upper() != "ID"]
         non_id_col_list = ", ".join(f'"{c}"' for c in non_id_cols)
 
+        # Time temp table creation
+        start = time.time()
         cursor.execute(
             f'CREATE TEMP TABLE "{temp_table}" AS SELECT {non_id_col_list} FROM "{table}" WHERE FALSE'  # noqa: S608
         )
+        worker_times["temp_table_ops"] += time.time() - start
 
-        # 1) COPY data into staging
+        # 1) Time CSV generation and COPY data into staging
+        start = time.time()
         buf = io.StringIO()
         batch.write_csv(buf)
         buf.seek(0)
+        worker_times["csv_gen"] += time.time() - start
+        
+        start = time.time()
         cursor.copy_expert(f'COPY "{temp_table}" ({non_id_col_list}) FROM STDIN WITH CSV HEADER', buf)
+        worker_times["copy_data"] += time.time() - start
 
-        # 2) INSERT FROM staging into real table
+        # 2) Time INSERT FROM staging into real table
+        start = time.time()
         insert_sql = (
             f'INSERT INTO "{table}" ({non_id_col_list}) SELECT {non_id_col_list} FROM "{temp_table}" '  # noqa: S608
             + on_conflict
         )
         cursor.execute(insert_sql)
+        worker_times["insert"] += time.time() - start
 
-        # 3) Optionally retrieve IDs
+        # 3) Time ID fetching
         id_df = None
         if returning_id:
+            start = time.time()
             buf_out = io.StringIO()
             # Build SELECT clause with explicit aliases
             select_list = ", ".join(['r."ID" AS "ID"', *[f't."{c}" AS "{c}"' for c in pks]])
@@ -948,9 +1027,12 @@ def _copy_worker_static(
             cursor.copy_expert(copy_sql, buf_out)
             buf_out.seek(0)
             id_df = pl.read_csv(buf_out)
+            worker_times["fetch_ids"] += time.time() - start
 
-        # 4) Clean up temp table
+        # 4) Time cleanup
+        start = time.time()
         cursor.execute(f'DROP TABLE IF EXISTS "{temp_table}"')
+        worker_times["temp_table_ops"] += time.time() - start
 
         # Commit this worker's transaction
         raw_conn.commit()
@@ -961,7 +1043,7 @@ def _copy_worker_static(
         print(f"Worker error for offset {offset}")
         raise
     else:
-        return (batch_size, id_df)
+        return (batch_size, id_df, worker_times)
     finally:
         cursor.close()
         raw_conn.close()
